/*-----------------------------------------------------------------------------
align_ops.c

Shyam Prabhakar Aug 08 2003
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "tree_strucs.h"
#include "align_strucs.h"
#include "read_file.h"
#include "coord_map_strucs.h"

#define SEQ_ALLOC_SIZ 1000
#define BUF_SIZ 100
#define MAP_ALLOC_SIZ 20
#define CONTIG_END_SLOP 20


void list_chars(ALIGNMENT *algnmnt_ptr)
{

  int nrow, i;
  char **matrix, *seq;
  unsigned char c;
  unsigned int char_vector[256] = {0};

  if ( !algnmnt_ptr )  return;

  nrow   = algnmnt_ptr->nrow;
  matrix = algnmnt_ptr->matrix;

  for (i = 0; i < nrow; i++)
    {
      seq = matrix[i];
      while (c = *seq++)
	++char_vector[c];
    }

  for (i = 0; i < 256; i++)
    if (char_vector[i])
      printf("%c, ", i);

}


void print_maps(ALIGNMENT *algnmnt_ptr)
{

  COORD_MAP *maps;
  int row, nrow;
  int i, map_siz, *algn_coords, *seq_coords;
  char **seq_names;

  if ( !algnmnt_ptr)  return;

  if ( !(maps = algnmnt_ptr->coord_maps) )  return;
  
  seq_names = algnmnt_ptr->seq_names;
  nrow      = algnmnt_ptr->nrow;

  for (row = 0; row < nrow; row++)
    {
      printf("%s\n", seq_names[row]);

      if (map_siz = maps[row].map_siz)
	{
	  algn_coords = maps[row].algn_coords;
	  seq_coords  = maps[row].seq_coords;

	  for (i = 0; i < map_siz; i++)
	    printf("%d=%d, ", *algn_coords++, *seq_coords++);
	  printf("\n\n");
	}
    }

}


void print_alignment(ALIGNMENT *algnmnt_ptr, int col_start, int ncol_prnt)
{

  char **matrix;
  char **seq_names;
  int i, ncol;

  if ( !algnmnt_ptr)  return;

  matrix    = algnmnt_ptr->matrix;
  seq_names = algnmnt_ptr->seq_names;

  ncol = algnmnt_ptr->ncol;
  printf("nrow = %d, ncol = %d\n", algnmnt_ptr->nrow, ncol);

  if (col_start < ncol)
    for (i = 0; i < algnmnt_ptr->nrow; i++)
      {
	printf("%s\n", seq_names[i]);
	printf("%.*s\n", ncol_prnt, matrix[i] + col_start);
      }

}


char *translate_and_map(char *algn_seq, int algn_len, char *translator,
			COORD_MAP *map_ptr)
{

  int i, in_gap;
  unsigned char c, *s = algn_seq;
  int map_alloc_siz, map_siz, *algn_coords, *seq_coords, seq_len;

  if (!algn_len)
    {
      map_ptr->map_siz = 0;
      return algn_seq;
    }

  map_alloc_siz = MAP_ALLOC_SIZ > 2 ? MAP_ALLOC_SIZ : 2;

  algn_coords = (int *) malloc(map_alloc_siz * sizeof(int));
  seq_coords  = (int *) malloc(map_alloc_siz * sizeof(int));

  map_siz = 1;
  seq_len = 0;
  algn_coords[0] = seq_coords[0] = -1;

  if ( (c = *s) == '-' )
    {
      s++;
      in_gap = 1;
    }
  else
    {
      *s++ = translator[c];
      in_gap = 0;

      algn_coords[1] = seq_coords[1] = 0;
      map_siz++;
      seq_len++;
    }

  for (i = 1; i < algn_len; i++)
    {
      c = *s;

      if (in_gap)
	if (c == '-') s++;
	else
	  {
	    if (*--s == '<') *s++ = '*';  /* gap of length 1 (start & end) */
	    else             *s++ = '>';  /* end gap */

	    *s++ = translator[c];
	    in_gap = 0;

	    if ( !(map_siz % map_alloc_siz) )
	      {
		algn_coords = (int *)
		  realloc(algn_coords, (map_siz + map_alloc_siz)*sizeof(int) );
		seq_coords  = (int *)
		  realloc(seq_coords,  (map_siz + map_alloc_siz)*sizeof(int) );
	      }

	    algn_coords[map_siz]  = i;
	    seq_coords[map_siz++] = seq_len++;
	  }
      else
	if (c == '-')
	  {
	    *s++ = '<';  /* start gap */
	    in_gap = 1;
	  }
	else
	  {
	    *s++ = translator[c];
	    seq_len++;
	  }
    }

  if ( !(map_siz % map_alloc_siz) )
    {
      algn_coords = (int *) realloc(algn_coords, (map_siz + 1)*sizeof(int) );
      seq_coords  = (int *) realloc(seq_coords,  (map_siz + 1)*sizeof(int) );
    }

  algn_coords[map_siz]  = algn_len;
  seq_coords[map_siz++] = seq_len;

  map_ptr->map_siz     = map_siz;
  map_ptr->algn_coords = algn_coords;
  map_ptr->seq_coords  = seq_coords;

  return algn_seq;

}


int store_seq_name(char **seq_names, char *seq_name, int nrow)
{

  int i;

  for (i = 0; i < nrow-1; i++)
    if ( !strcmp(seq_name, seq_names[i]) )
      {
	fprintf(stderr, "Two sequences have the name: %s!\n", seq_name);
	return 0;
      }

  seq_names[nrow-1] = seq_name;
  return 1;

}


void crash_read_align_fasta(FILE *fp, char *seq, char **matrix1,
			    COORD_MAP *maps, int nrow1, char **matrix2,
			    int nrow2, ALIGNMENT *algnmnt_ptr)
{

  int i;

  if (fp)  fclose(fp);

  if (seq)  free(seq);

  if (algnmnt_ptr)  free(algnmnt_ptr);

  if (nrow1)
    {
      for (i = 0; i < nrow1; i++)
	{
	  free(matrix1[i]);
	  free(maps[i].algn_coords);
	  free(maps[i].seq_coords);
	}

      free(matrix1);
      free(maps);
    }

  if (nrow2)
    {
      for (i = 0; i < nrow2; i++)
	free(matrix2[i]);

      free(matrix2);
    }

}


void init_lett_translator(char *transl)
{

  int i;

  for (i = 0; i < 256; i++)
    transl[i] = 'N';

  transl['a'] = transl['A'] = 'A';
  transl['g'] = transl['G'] = 'G';
  transl['c'] = transl['C'] = 'C';
  transl['t'] = transl['T'] = 'T';

  transl['-'] = '-';

}


ALIGNMENT *read_align_fasta(char *fname)
{

  FILE *fp = 0;
  ALIGNMENT *algnmnt_ptr = 0;
  int alloc_siz = SEQ_ALLOC_SIZ;
  int nrow, ncol, algn_len, seq_buf_siz, buf_str_len;
  char buffer[BUF_SIZ + 1], c;
  char *seq = 0, *seq_name, **seq_names, **matrix, *end_ptr;
  char translator[256];
  COORD_MAP *maps;

  if ( !(fp = fopen(fname, "r")) )
    {
      fprintf(stderr, "couldn't open file %s\n!", fname);
      return 0;
    }

  /* read until start of first sequence */
  while ( fgets(buffer, BUF_SIZ + 1, fp) && *buffer != '>');

  if (*buffer != '>')
    {
      fprintf(stderr, "no sequences in fasta file %s\n!", fname);
      fclose(fp);
      return 0;
    }

  if (alloc_siz < BUF_SIZ)  alloc_siz = BUF_SIZ;

  algnmnt_ptr = (ALIGNMENT *) calloc(1, sizeof(ALIGNMENT));

  nrow      = 0;
  ncol      = 0;
  seq_names = (char **) malloc(0);
  matrix    = (char **) malloc(0);
  maps      = (COORD_MAP *) malloc(0);

  init_lett_translator(translator);

  do
    {

      /* ignore strange lines beginning with '\0' */
      if ( !(buf_str_len = strlen(buffer)) )  continue;

      if (*buffer == '>')
	{
	  
	  if (nrow)
	    {

	      if (ncol)
		{
		  if (algn_len != ncol)
		    {
		      fprintf(stderr, "lengths of alignment rows differ!\n");
		      fprintf(stderr, "(algn_len = %d, ncol = %d)\n",
			      algn_len, ncol);
		      crash_read_align_fasta(fp, seq, matrix, maps, nrow-1,
					     seq_names, nrow, algnmnt_ptr);
		      return 0;
		    }
		}
	      else
		ncol = algn_len;

	      matrix[nrow-1] =
		translate_and_map(seq, algn_len, translator, maps + nrow - 1);

	    }

	  ++nrow;

	  seq_names = (char **) realloc(seq_names, nrow*sizeof(char *) );

	  if ( !(seq_name = read_first_word(buffer, 'l')) )
	    {
	      seq_name = malloc(20*sizeof(char));
	      sprintf(seq_name, "seq_%d", nrow);

	      seq_names[nrow-1] = seq_name;
	    }
	  else if ( !store_seq_name(seq_names, seq_name, nrow) )
	    {
	      crash_read_align_fasta(fp, 0, matrix, maps, nrow-1, seq_names,
				     nrow, algnmnt_ptr);
	      return 0;
	    }

	  matrix = (char **) realloc(matrix, nrow*sizeof(char *) );
	  maps   = (COORD_MAP *) realloc(maps, nrow*sizeof(COORD_MAP) );

	  algn_len     = 0;
	  seq_buf_siz = alloc_siz;
	  seq         = (char *) calloc(alloc_siz + 1, sizeof(char) );

	  /* if entire line did not fit in buffer, advance fp to end of line */
	  if ( buffer[buf_str_len - 1] != '\n' )
	    while ( (c = fgetc(fp)) != '\n' && c != EOF )  ;

	}
      else
	{

	  /* ignore trailing whitespace characters */
	  end_ptr = buffer + buf_str_len;
	  while ( isspace(*--end_ptr) && --buf_str_len )  ;

	  if (algn_len + buf_str_len > seq_buf_siz)
	    {
	      seq_buf_siz += alloc_siz;
	      seq = (char *) realloc(seq, (seq_buf_siz + 1)*sizeof(char));
	    }

	  strncpy(seq + algn_len, buffer, buf_str_len);
	  algn_len += buf_str_len;

	}

    } while ( fgets(buffer, BUF_SIZ + 1, fp) );

  if (nrow < 2)
    {
      fprintf(stderr, "less than two sequences in alignment ... abort!\n");
      crash_read_align_fasta(fp, seq, matrix, maps, nrow-1, seq_names, nrow,
			     algnmnt_ptr);
      return 0;
    }

  if (algn_len != ncol)
    {
      fprintf(stderr, "lengths of aligned seqs differ!\n");
      fprintf(stderr, "(algn_len = %d, ncol = %d)\n", algn_len, ncol);

      crash_read_align_fasta(fp, seq, matrix, maps, nrow-1, seq_names, nrow,
			     algnmnt_ptr);
      return 0;
    }

  fclose(fp);

  matrix[nrow-1] = translate_and_map(seq, algn_len, translator, maps+nrow-1);

  algnmnt_ptr->nrow       = nrow;
  algnmnt_ptr->ncol       = ncol;
  algnmnt_ptr->seq_names  = seq_names;
  algnmnt_ptr->matrix     = matrix;
  algnmnt_ptr->coord_maps = maps;

/*   print_alignment(algnmnt_ptr, 0, 60); */
/*   print_maps(algnmnt_ptr); */

  return algnmnt_ptr;

}


ALIGNMENT *read_align(char *fname, char *fmt)
{

  ALIGNMENT *algnmnt_ptr = 0;

  if ( !fmt || !strcmp(fmt, "fasta") )  /* fasta is default format */
    algnmnt_ptr = read_align_fasta(fname);
  else
    fprintf(stderr, "unrecognised alignment format: %s!\n", fmt);

/*   list_chars(algnmnt_ptr); */

  return algnmnt_ptr;

}


void identify_gap_chars(int is_gap[256])
{

  is_gap[(unsigned char) '-'] = is_gap[(unsigned char) '<'] =
  is_gap[(unsigned char) '>'] = is_gap[(unsigned char) '*'] = 1;

}


void proc_non_AGCT(ALIGNMENT *algnmnt_ptr)
{

  int nrow, ncol, r, c;
  int *starts, *ends, in_N_seg, N_seg_len, num_N_seg;
  int *col_AGCT_distro, *col_num_AGCT, *row_num_AGCT;
  int *seq_lens, seq_len, seq_AGCT;
  char **matrix, ch;
  unsigned char *seq;
  int is_gap[256] = {0};
  SEG_LIST *row_N_segs;

  nrow   = algnmnt_ptr->nrow;
  ncol   = algnmnt_ptr->ncol;
  matrix = algnmnt_ptr->matrix;

  identify_gap_chars(is_gap);

  col_AGCT_distro = (int *) calloc(nrow+1, sizeof(int));
  col_num_AGCT    = (int *) calloc(ncol, sizeof(int));
  row_num_AGCT    = (int *) calloc(nrow, sizeof(int));
  seq_lens        = (int *) calloc(nrow, sizeof(int));
  row_N_segs      = (SEG_LIST *) calloc(nrow, sizeof(SEG_LIST));

  /* inefficient. Should use coord_maps information. -Shyam */
  for (r = 0; r < nrow; r++)
    {
      seq = matrix[r];
      seq_AGCT = seq_len = ncol;

      in_N_seg = N_seg_len = num_N_seg = 0;
      starts = (int *) malloc(0);
      ends   = (int *) malloc(0);

      for (c = 0; c < ncol && (ch = *seq++); c++)
	if (in_N_seg)
	  if ( ch == 'N' )
	    {
	      --col_num_AGCT[c], --seq_AGCT;
	      N_seg_len++;
	    }
	  else
	    {
	      if ( is_gap[ch] )
		--col_num_AGCT[c], --seq_AGCT, --seq_len;

	      in_N_seg = 0;
	      ends[num_N_seg++] = c - 1;
	    }
	else
	  if ( is_gap[ch] )
	    --col_num_AGCT[c], --seq_AGCT, --seq_len;
	  else if ( ch == 'N' )
	    {
	      --col_num_AGCT[c], --seq_AGCT;

	      if ( !(num_N_seg % MAP_ALLOC_SIZ) )
		{
		  starts = (int *)
		    realloc(starts, (num_N_seg + MAP_ALLOC_SIZ)*sizeof(int) );
		  ends   = (int *)
		    realloc(ends,   (num_N_seg + MAP_ALLOC_SIZ)*sizeof(int) );
		}

	      N_seg_len = in_N_seg = 1;
	      starts[num_N_seg] = c;
	    }

      if (in_N_seg)  ends[num_N_seg++] = ncol - 1;

      row_num_AGCT[r]    = seq_AGCT;
      seq_lens[r]        = seq_len;

      row_N_segs[r].nseg   = num_N_seg;
      row_N_segs[r].starts = starts;
      row_N_segs[r].ends   = ends;

    }

  for (c = 0; c < ncol; c++)
    ++col_AGCT_distro[col_num_AGCT[c] += nrow];

  algnmnt_ptr->col_AGCT_distro = col_AGCT_distro;
  algnmnt_ptr->col_num_AGCT    = col_num_AGCT;
  algnmnt_ptr->row_num_AGCT    = row_num_AGCT;
  algnmnt_ptr->seq_lens        = seq_lens;
  algnmnt_ptr->row_N_segs      = row_N_segs;

}


/*-----------------------------------------------------------------------------
IMPORTANT - tree_algn_ptr mostly points to chunks of memory malloc-ed by other
structs, so it should be freed as follows: 

free(tree_algn_ptr->matrix);
free(tree_algn_ptr);
-----------------------------------------------------------------------------*/
ALIGNMENT *get_tree_align(PHYLO_TREE *phy_tree_ptr, ALIGNMENT *inp_algn_ptr)
{

  ALIGNMENT *tree_algn_ptr;
  int leaf, j;
  int nleaf, nrow;
  char **leaves, **seq_names;
  char **matrix, **tree_matrix;
  COORD_MAP *maps, *tree_maps;

  if ( !inp_algn_ptr )
    {
      fprintf(stderr, "empty input alignment!\n");
      return 0;
    }

  nleaf = phy_tree_ptr->nleaf;
  nrow  = inp_algn_ptr->nrow;

  matrix = inp_algn_ptr->matrix;
  maps   = inp_algn_ptr->coord_maps;

  tree_matrix = (char **) calloc(nleaf, sizeof(char *));
  tree_maps   = (COORD_MAP *) calloc(nleaf, sizeof(COORD_MAP));

  leaves    = phy_tree_ptr->leaves;
  seq_names = inp_algn_ptr->seq_names;

  for (leaf = 0; leaf < nleaf; leaf++)
    {
      j = -1;
      while ( ++j < nrow && strcmp(leaves[leaf], seq_names[j]) );

      if (j == nrow)
	{
	  fprintf(stderr, "No alignment found for tree leaf: %s!\n",
		  leaves[leaf]);

	  free(tree_matrix);
	  free(tree_maps);

	  return 0;
	}
      else
	{
	  tree_matrix[leaf] = matrix[j];
	  tree_maps[leaf]   = maps[j];
	}

    }

  tree_algn_ptr = (ALIGNMENT *) malloc(sizeof(ALIGNMENT));

  tree_algn_ptr->nrow       = nleaf;
  tree_algn_ptr->ncol       = inp_algn_ptr->ncol;
  tree_algn_ptr->seq_names  = leaves;
  tree_algn_ptr->matrix     = tree_matrix;
  tree_algn_ptr->coord_maps = tree_maps;
  tree_algn_ptr->tree       = phy_tree_ptr;

  proc_non_AGCT(tree_algn_ptr);

 /*  print_alignment(tree_algn_ptr, 0, 150); */

  return tree_algn_ptr;

}
