/*-----------------------------------------------------------------------------
coord_map.c

Shyam Prabhakar Oct 22 2003
-----------------------------------------------------------------------------*/

#include <stdio.h>

#include "coord_map_strucs.h"


/* NOTE: it's kind of dangerous to assume that just because the map ptr is the
   same as in the last function call, the contents of the pointer are the same,
   but what the hell. This speeds the program up a little. */

int align_coord(int seq_coord, COORD_MAP *map_ptr)
{

  int map_siz, *algn_anchors, *seq_anchors;
  int mid_anchor, mid_idx, l_anchor_idx, u_anchor_idx;

  map_siz      = map_ptr->map_siz;
  algn_anchors = map_ptr->algn_coords;
  seq_anchors  = map_ptr->seq_coords;

  l_anchor_idx = 0;
  u_anchor_idx = map_siz-1;

  while (l_anchor_idx + 1 < u_anchor_idx)
    {

      mid_idx = (l_anchor_idx + u_anchor_idx)/2;

      mid_anchor = seq_anchors[mid_idx];

      if (seq_coord < mid_anchor)  u_anchor_idx = mid_idx;
      else                         l_anchor_idx = mid_idx;

    }

  return algn_anchors[l_anchor_idx] - seq_anchors[l_anchor_idx] + seq_coord;  

}


int seq_coord(int align_coord, char dirn, COORD_MAP *map_ptr)
{

  int map_siz, *algn_anchors, *seq_anchors;
  int mid_anchor, mid_idx, l_anchor_idx, u_anchor_idx;
  int offset, max_offset, coord;

  map_siz      = map_ptr->map_siz;
  algn_anchors = map_ptr->algn_coords;
  seq_anchors  = map_ptr->seq_coords;

  l_anchor_idx = 0;
  u_anchor_idx = map_siz-1;
  
  while (l_anchor_idx + 1 < u_anchor_idx)
    {

      mid_idx    = (l_anchor_idx + u_anchor_idx)/2;

      mid_anchor = algn_anchors[mid_idx];

      if (align_coord < mid_anchor)  u_anchor_idx = mid_idx;
      else                           l_anchor_idx = mid_idx;

    }

  offset     = align_coord - algn_anchors[l_anchor_idx];
  max_offset = seq_anchors[u_anchor_idx] - seq_anchors[l_anchor_idx] - 1;

  /* if alignment coordinate corresponds to a gap position, map it to
     the sequence coordinate immediately to the left or right of the gap */
  if (offset > max_offset)
    if (dirn == '<')
      offset = max_offset;
    else
      offset = max_offset + 1;

  /* returning a coordinate of -1 could be dangerous... */
  if ( (coord = seq_anchors[l_anchor_idx] + offset) == -1 )
    return 0;

  return coord;

}


/* generate a map assuming no indels in an aligned sequence
   of length len, i.e., a map of the sequence to itself */
COORD_MAP *self_map(int len)
{

  COORD_MAP *map_ptr = (COORD_MAP *) malloc(sizeof(COORD_MAP));

  map_ptr->map_siz = 3;
  map_ptr->algn_coords = (int *) calloc(3, sizeof(int));
  map_ptr->seq_coords  = (int *) calloc(3, sizeof(int));

  map_ptr->algn_coords[0] = map_ptr->seq_coords[0] = -1;
  map_ptr->algn_coords[2] = map_ptr->seq_coords[2] = len;

  return map_ptr;

}
