/*-----------------------------------------------------------------------------
prnt_gumby_matlab.c

Shyam Prabhakar Oct 21 2003
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>

#include "scor_strucs.h"
#include "hit_strucs.h"
#include "annot_strucs.h"
#include "align_strucs.h"
#include "classify_annots.h"
#include "coord_map.h"
#include "arg_struc.h"


void prnt_annots(GFF_ANNOT *annots, char **seq_names, int nrow,
		 COORD_MAP *map_ptr, FILE *fp)
{

  int i, row, nfeat, annot_code;
  int feat_start, feat_end;
  char *feat_name;
  GFF_FEAT *feats;

  for (row = 0; row < nrow; row++)
    if (nfeat = annots[row].num_feat)
      {

	feats = annots[row].feats;

	fprintf(fp, "%s_annots = [\n", seq_names[row]);
	for (i = 0; i < nfeat; i++)
	  {

	    /* project coordinates */
	    feat_start = seq_coord(feats[i].algn_start, '>', map_ptr);
	    feat_end   = seq_coord(feats[i].algn_end, '<', map_ptr);

/* 	    if ( feat_end >= feat_start ) */
	    if ( feat_end >= 0 )
	      {

		feat_name = feats[i].feature;
		if ( is_exon(feat_name) || is_utr(feat_name) )  annot_code = 1;
		else if ( is_cds(feat_name)  )  annot_code = 2;
		else                            annot_code = 0;

		/* one is added to make original seq.
		   coords. one-based in the output */
		fprintf(fp, "%d %d %d %d %d;\n", feats[i].start+1,
			feats[i].end+1,	feat_start, feat_end, annot_code);

	      }

	  }
	fprintf(fp, "];\n\n");

      }

}


void prnt_hits(HIT1D *hits, int nhit, COORD_MAP *map_ptr, FILE *fp)
{

  int i, ah_start, ah_end, h_start, h_end;

  fprintf(fp, "hits = [hits;\n");
  for (i = 0; i < nhit; i++)
    {

      ah_start = hits[i].start;
      ah_end   = ah_start + hits[i].len - 1;

      h_start = seq_coord(ah_start, '>', map_ptr);
      h_end   = seq_coord(ah_end,   '<', map_ptr);

      if ( h_end >= h_start )
	{
	  fprintf(fp, "%d %d ", h_start, h_end);
	  fprintf(fp, "%d %g;\n", hits[i].scor, hits[i].pval);
	}
      else
	{
	  fprintf(stderr, "Hit deleted in base sequence: "
		  "align_start = %d, align_end = %d\n", ah_start, ah_end);
	}

    }
  fprintf(fp, "];\n\n");

}


void prnt_scores(int *scores, COORD_MAP *map_ptr, FILE *fp)
{

  int i, m, map_siz, *algn_coords, *seq_coords;
  int seg_start, seg_end;

  map_siz     = map_ptr->map_siz;
  algn_coords = map_ptr->algn_coords;
  seq_coords  = map_ptr->seq_coords;

  fprintf(fp, "scores = [\n");

  for (m = 1; m < map_siz - 1; m++)
    {

      seg_start = algn_coords[m];  /* zero-based */
      seg_end   = seg_start + seq_coords[m + 1] - seq_coords[m];  /* 1-based */

      for (i = seg_start; i < seg_end; fprintf(fp, "%d ", scores[i++]) )  ;

    }

  fprintf(fp, "];\n\n");

}


void prnt_gumby_matlab(ALIGN_SCOR *scor_ptr, HIT1D *hits, int nhit,
			ALIGNMENT *algnmnt_ptr, int proj, FILE *fp, ARGS args)
{

  int nrow, scor_len;
  char **seq_names;
  static COORD_MAP *map_ptr;

  /* if map_ptr = 0, it's a signal that most of the file has already
     been written, and only the hits need to be written now. */

  nrow      = algnmnt_ptr->nrow;
  seq_names = algnmnt_ptr->seq_names;

  if (!map_ptr)
    {

      if ( proj >=0 && proj < nrow )
	{
	  map_ptr  = algnmnt_ptr->coord_maps + proj;
	  scor_len = algnmnt_ptr->seq_lens[proj];

	  fprintf(fp, "%%%s\n\n", args.invocation);
	  fprintf(fp, "proj = '%s';\n", seq_names[proj]);
	}
      else
	{
	  map_ptr = self_map(scor_len = scor_ptr->len);

	  fprintf(fp, "%%%s\n\n", args.invocation);
	  fprintf(fp, "proj = [];\n");
	}

      fprintf(fp, "len = %d;\n", scor_len);
/*       prnt_scores(scor_ptr->scor, map_ptr, fp); */
      prnt_annots(algnmnt_ptr->annots, seq_names, nrow, map_ptr, fp);

      fprintf(fp, "hits = [];\n\n");

    }

  prnt_hits(hits, nhit, map_ptr, fp);

}
